from laktory.models import Catalog


catalog = Catalog(
        name="dev",
        grants=[
            {"principal": "account users", "privileges": ["USE_CATALOG", "USE_SCHEMA"]}
        ],
        schemas=[
            {
                "name": "engineering",
                "grants": [
                    {"principal": "domain-engineering", "privileges": ["SELECT"]}
                ],
            },
            {
                "name": "sources",
                "volumes": [
                    {
                        "name": "landing",
                        "volume_type": "EXTERNAL",
                        "grants": [
                            {
                                "principal": "account users",
                                "privileges": ["READ_VOLUME"],
                            },
                            {
                                "principal": "role-metastore-admins",
                                "privileges": ["WRITE_VOLUME"],
                            },
                        ],
                    },
                ],
            },
        ],
    )


def test_model():
    assert catalog.name == "dev"
    assert catalog.full_name == "dev"
    assert catalog.full_name == "dev"


if __name__ == "__main__":
    test_model()



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.userrole import UserRole
from laktory.models.databricks.groupmember import GroupMember


class User(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks user

    Attributes
    ----------
    disable_as_user_deletion:
        If `True` user is disabled instead of delete when the resource is deleted
    display_name:
        Display name for the user
    id:
        Id of the user. Generally used when the user is externally managed
        with an identity provider such as Azure AD, Okta or OneLogin.
    group_ids:
        List of the group ids that the user should be member of.
    roles:
        List of roles assigned to the user e.g. ("account_admin")
    workspace_access
        When `True`, the user is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    u = models.User(
        user_name="john.doe@okube.ai",
        display_name="John Doe",
        group_ids=[
            "${resources.group-role-engineer.id}",
            "${resources.group-domain-finance.id}",
        ],
        roles=["account_admin"],
    )
    ```
    """

    disable_as_user_deletion: bool = False
    display_name: str = None
    id: Union[str, None] = None
    group_ids: list[str] = []
    roles: list[str] = []
    user_name: str
    workspace_access: bool = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.user_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - user roles
        - user group members
        """
        resources = []
        for role in self.roles:
            resources += [
                UserRole(
                    user_id=f"${{resources.{self.resource_name}.id}}",
                    role=role,
                )
            ]

        # Group Member
        for group_id in self.group_ids:
            resources += [
                GroupMember(
                    group_id=group_id,
                    member_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:User"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.User

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["groups", "roles", "id", "group_ids"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_user"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.metastoreassignment import MetastoreAssignment
from laktory.models.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.grants.metastoregrant import MetastoreGrant
from laktory.models.databricks.grants import Grants


class Metastore(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Metastore

    Attributes
    ----------
    cloud:
        todo
    created_at:
        todo
    created_by:
        todo
    default_data_access_config_id:
        todo
    delta_sharing_organization_name:
        The organization name of a Delta Sharing entity. This field is used for
         Databricks to Databricks sharing. Once this is set it cannot be
        removed and can only be modified to another valid value. To delete
        this value please taint and recreate the resource.
    delta_sharing_recipient_token_lifetime_in_seconds:
        Required along with `delta_sharing_scope`. Used to set expiration
        duration in seconds on recipient data access tokens.
        Set to 0 for unlimited duration.
    delta_sharing_scope:
        Required along with delta_sharing_recipient_token_lifetime_in_seconds.
         Used to enable delta sharing on the metastore.
        Valid values: INTERNAL, INTERNAL_AND_EXTERNAL.
    force_destroy:
        Destroy metastore regardless of its contents.
    global_metastore_id:
        todo
    grants:
        List of grants operating on the metastore
    metastore_id:
        todo
    name:
        Name of metastore.
    owner:
        Username/groupname/sp application_id of the metastore owner.
    region:
        The region of the metastore
    storage_root:
        Path on cloud storage account, where managed databricks.Table are
        stored. Change forces creation of a new resource. If no storage_root is
        defined for the metastore, each catalog must have a storage_root
        defined.
    storage_root_credential_id:
        todo
    updated_at:
        todo
    updated_by:
        todo
    workspace_assignments:
        List of workspace to which metastore is assigned to

    Examples
    --------
    ```py
    ```
    """

    cloud: str = None
    created_at: int = None
    created_by: str = None
    data_accesses: list[MetastoreDataAccess] = None
    default_data_access_config_id: str = None
    delta_sharing_organization_name: str = None
    delta_sharing_recipient_token_lifetime_in_seconds: int = None
    delta_sharing_scope: str = None
    force_destroy: bool = None
    global_metastore_id: str = None
    grants: list[MetastoreGrant] = None
    metastore_id: str = None
    name: str = None
    owner: str = None
    region: str = None
    storage_root: str = None
    storage_root_credential_id: str = None
    updated_at: int = None
    updated_by: str = None
    workspace_provider: str = None
    workspace_assignments: list[MetastoreAssignment] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace assignments
        - grants
        """
        resources = []

        if self.workspace_assignments:
            for a in self.workspace_assignments:
                a.metastore_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]

        if self.grants:
            resources += Grants(
                resource_name=f"grants-{self.resource_name}",
                metastore=f"${{resources.{self.resource_name}.id}}",
                grants=[
                    {"principal": g.principal, "privileges": g.privileges}
                    for g in self.grants
                ],
                options={"provider": self.workspace_provider},
            ).core_resources

        if self.data_accesses:
            for data_access in self.data_accesses:
                data_access.metastore_id = f"${{resources.{self.resource_name}.id}}"
                # TODO: Add dependency on grants
                _core_resources = data_access.core_resources
                for r in _core_resources[1:]:
                    if r.options.provider is None:
                        r.options.provider = self.workspace_provider
                resources += _core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Metastore"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Metastore

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "workspace_assignments",
            "grants",
            "data_accesses",
            "workspace_provider",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_metastore"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes + ["cloud"]



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.databricks.grants import Grants
from laktory.models.grants.storagecredentialgrant import StorageCredentialGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MetastoreDataAccessAwsIamRole(BaseModel):
    """
    Databricks Metastore Data Access AWS IAM Role

    Attributes
    ----------
    role_arn:
        todo
    external_id:
        todo
    unity_catalog_iam_arn:
        todo
    """

    role_arn: str = None
    external_id: str = None
    unity_catalog_iam_arn: str = None


class MetastoreDataAccessAzureManagedIdentity(BaseModel):
    """
    Databricks Metastore Data Access Azure Managed Identity

    Attributes
    ----------
    access_connector_id:
        todo
    credential_id:
        todo
    managed_identity_id:
        todo
    """

    access_connector_id: str = None
    credential_id: str = None
    managed_identity_id: str = None


class MetastoreDataAccessAzureServicePrincipal(BaseModel):
    """
    Databricks Metastore Data Access Azure Service Principals

    Attributes
    ----------
    application_id:
        todo
    client_secret:
        todo
    directory_id:
        todo
    """

    application_id: str = None
    client_secret: str = None
    directory_id: str = None


class MetastoreDataAccessDatabricksGcpServiceAccount(BaseModel):
    """
    Databricks Metastore Data Access GCP Service Account Key

    Attributes
    ----------
    email:
        todo
    private_key:
        todo
    private_key_id:
        todo
    """

    credential_id: str = None
    email: str = None


class MetastoreDataAccessGcpServiceAccountKey(BaseModel):
    """
    Databricks Metastore Data Access GCP Service Account Key

    Attributes
    ----------
    email:
        todo
    private_key:
        todo
    private_key_id:
        todo
    """

    email: str = None
    private_key: str = None
    private_key_id: str = None


class MetastoreDataAccess(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Metastore Data Access

    Attributes
    ----------
    aws_iam_role:
        AWS IAM role specifications
    azure_managed_identity:
        Azure Managed Identity specifications
    azure_service_principal:
        Azure Service Principal specifications
    comment:
        Comment
    databricks_gcp_service_account:
        Databricks GCP service account specifications
    force_destroy:
        Force resource deletion even if not empty
    force_update:
        Force resource update even if not empty
    grants:
        List of grants operating on the data access
    gcp_service_account_key:
        GCP service account key specifications
    is_default:
        Whether to set this credential as the default for the metastore.
    metastore_id:
        Metastore id
    name:
        Data Access name
    owner:
        Owner
    read_only:
        Read only
    skip_validation:
        Skip Validation

    Examples
    --------
    ```py
    ```
    """

    aws_iam_role: MetastoreDataAccessAwsIamRole = None
    azure_managed_identity: MetastoreDataAccessAzureManagedIdentity = None
    azure_service_principal: MetastoreDataAccessAzureServicePrincipal = None
    comment: str = None
    databricks_gcp_service_account: MetastoreDataAccessDatabricksGcpServiceAccount = (
        None
    )
    force_destroy: bool = None
    force_update: bool = None
    gcp_service_account_key: MetastoreDataAccessGcpServiceAccountKey = None
    grants: list[StorageCredentialGrant] = None
    is_default: bool = None
    metastore_id: str = None
    name: str = None
    owner: str = None
    read_only: bool = None
    skip_validation: bool = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - storage credential grants
        """
        resources = []

        # Catalog grants
        if self.grants:
            grant = Grants(
                resource_name=f"grants-{self.resource_name}",
                storage_credential=self.name,
                grants=[
                    {"principal": g.principal, "privileges": g.privileges}
                    for g in self.grants
                ],
            )
            resources += [grant]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MetastoreDataAccess"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.MetastoreDataAccess

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_metastore_data_access"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from pyspark.sql.dataframe import DataFrame

from laktory.spark.dataframe.schema_flat import schema_flat
from laktory.spark.dataframe.has_column import has_column
from laktory.spark.dataframe.show_string import show_string


# DataFrame Extensions
DataFrame.schema_flat = schema_flat
DataFrame.has_column = has_column
DataFrame.show_string = show_string

# Spark Connect DataFrame Extensions
try:
    from pyspark.sql.connect.dataframe import DataFrame

    DataFrame.schema_flat = schema_flat
    DataFrame.has_column = has_column
    DataFrame.show_string = show_string
except:
    pass



from pydantic import model_validator
from typing import Union
from typing import Any
from settus import BaseSettings
from settus import Field
from settus import SettingsConfigDict


class Settings(BaseSettings):
    # model_config = SettingsConfigDict(
    #     keyvault_url="LAKTORY_KEYVAULT_URL"
    # )

    # CLI
    cli_raise_external_exceptions: bool = Field(
        False, alias="LAKTORY_CLI_RAISE_EXTERNAL_EXCEPTIONS"
    )

    # Configuration
    iac_backend: Union[str, None] = Field("pulumi", alias="LAKTORY_IAC_BACKEND")

    # Models
    camel_serialization: bool = Field(False)
    singular_serialization: bool = Field(False)

    # Azure
    lakehouse_sa_conn_str: Union[str, None] = Field(None, alias="LAKEHOUSE_SA_CONN_STR")

    # AWS
    aws_access_key_id: Union[str, None] = Field(None, alias="AWS_ACCESS_KEY_ID")
    aws_region: Union[str, None] = Field(None, alias="AWS_REGION")
    aws_secret_access_key: Union[str, None] = Field(None, alias="AWS_SECRET_ACCESS_KEY")

    # Databricks
    workspace_env: str = Field("dev", alias="LAKTORY_WORKSPACE_ENV")
    workspace_laktory_root: str = Field(
        "/.laktory/", alias="LAKTORY_WORKSPACE_LAKTORY_ROOT"
    )  # TODO: Review variable name
    workspace_landing_root: str = Field("", alias="LAKTORY_WORKSPACE_LANDING_ROOT")

    # Logging
    log_level: str = Field("INFO", alias="LAKTORY_LOG_LEVEL")

    # https://learn.microsoft.com/en-ca/azure/databricks/dev-tools/auth#general-host-token-and-account-id-environment-variables-and-fields
    # databricks_host: Union[str, None] = Field(None, alias="databricks_host")
    # databricks_token: Union[str, None] = Field(None, alias="databricks_token")
    # databricks_account_id: Union[str, None] = Field(None, alias="databricks_account_id")
    # databricks_warehouse_id: Union[str, None] = Field(None, alias="databricks_warehouse_id")
    databricks_host: Union[str, None] = Field(None)
    databricks_token: Union[str, None] = Field(None)
    databricks_account_id: Union[str, None] = Field(None)
    databricks_warehouse_id: Union[str, None] = Field(None)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        if self.workspace_landing_root == "":
            self.workspace_landing_root = (
                f"/Volumes/{self.workspace_env}/sources/landing/"
            )

        return self


settings = Settings()



from .basemodel import BaseModel
from .databricks import *
from .dataevent import DataEvent
from .dataeventheader import DataEventHeader
from .dataproducer import DataProducer
from .datasources import *
from .grants import *
from .providers import *
from .resources import *
from .sql import *
from .stacks import *



from pydantic import model_validator
from typing import Any
from typing import Literal
from typing import Union
from typing import Callable

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.models.datasources import EventDataSource
from laktory.models.datasources import TableDataSource
from laktory.models.sql.column import Column
from laktory.models.sql.tableaggregation import TableAggregation
from laktory.models.sql.tablejoin import TableJoin
from laktory.models.sql.tablewindowfilter import TableWindowFilter
from laktory.spark import DataFrame

logger = get_logger(__name__)


class TableBuilder(BaseModel):
    """
    Mechanisms for building a table from a source in the context of a data
    pipeline.

    Attributes
    ----------
    aggregation:
        Definition of the aggregation model following the joins and the source
        read.
    drop_columns:
        Columns to drop from the output dataframe
    drop_duplicates:
        If `True`, drop duplicated rows using table `primary_key`
    drop_source_columns:
        If `True`, drop columns from the source after read and only keep
        columns defined in the table and/or resulting from the joins.
    event_source:
        Definition of the event data source if applicable
    filter:
        Filter applied to the data source to keep only selected rows
    joins:
        Definition of the table(s) to be joined to the data source
    joins_post_aggregation
        Definition of the table(s) to be joined to the aggregated data
    layer:
        Layer in the medallion architecture
    pipeline_name:
        Name of the pipeline in which the table will be built
    selects:
        Columns to select from the output dataframe. A list of colum names or
        a map for renaming the columns.
    table_source:
        Definition of the table data source if applicable
    template:
        Key indicating which notebook to use for building the table in the
        context of a data pipeline.
    window_filter:
        Definition of rows filter based on a time spark window. Applied after
        joins.
    """

    aggregation: Union[TableAggregation, None] = None
    drop_columns: list[str] = []
    drop_duplicates: Union[bool, None] = None
    drop_source_columns: Union[bool, None] = None
    event_source: Union[EventDataSource, None] = None
    filter: Union[str, None] = None
    joins: list[TableJoin] = []
    joins_post_aggregation: list[TableJoin] = []
    layer: Literal["BRONZE", "SILVER", "SILVER_STAR", "GOLD"] = None
    pipeline_name: Union[str, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    table_source: Union[TableDataSource, None] = None
    template: Union[str, bool, None] = None
    window_filter: Union[TableWindowFilter, None] = None
    _columns_to_build = []
    _table: Any = None

    @model_validator(mode="after")
    def default_options(self) -> Any:
        """
        Sets default options like `drop_source_columns`, `drop_duplicates`,
        `template`, etc. based on `layer` value.
        """
        # Default values
        if self.layer == "BRONZE":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.layer == "SILVER":
            if self.drop_source_columns is None:
                self.drop_source_columns = True
            if self.drop_duplicates is not None:
                self.drop_duplicates = True

        if self.layer == "SILVER_STAR":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.layer == "GOLD":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.template is None:
            self.template = self.layer

        return self

    @property
    def source(self) -> BaseDataSource:
        """Selected data source"""
        if self.event_source is not None and self.event_source.name is not None:
            return self.event_source
        elif self.table_source is not None and self.table_source.name is not None:
            return self.table_source

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        if self.source is None:
            return False
        else:
            return self.source.is_cdc

    @property
    def columns(self) -> list[Column]:
        """List of columns"""
        return self._table.columns

    @property
    def timestamp_key(self) -> str:
        """Table timestamp key"""
        return self._table.timestamp_key

    @property
    def primary_key(self) -> str:
        """Table primary key"""
        return self._table.primary_key

    @property
    def has_joins(self) -> bool:
        """Joins defined flag"""
        return len(self.joins) > 0

    @property
    def has_joins_post_aggregation(self) -> bool:
        """Post-aggregation joins defined flag"""
        return len(self.joins_post_aggregation) > 0

    def _get_layer_columns(self, layer, df=None) -> list[columns]:
        from laktory.spark.dataframe import has_column

        cols = []

        if layer == "BRONZE":
            cols = [
                Column(
                    **{
                        "name": "_bronze_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif layer == "SILVER":
            if self.timestamp_key:
                cols += [
                    Column(
                        **{
                            "name": "_tstamp",
                            "type": "timestamp",
                            "spark_func_name": "coalesce",
                            "spark_func_args": [self.timestamp_key],
                        }
                    )
                ]

            if has_column(df, "_bronze_at"):
                cols += [
                    Column(
                        **{
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "spark_func_name": "coalesce",
                            "spark_func_args": ["_bronze_at"],
                        }
                    )
                ]

            cols += [
                Column(
                    **{
                        "name": "_silver_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif layer == "SILVER_STAR":
            cols = [
                Column(
                    **{
                        "name": "_silver_star_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif layer == "GOLD":
            cols = [
                Column(
                    **{
                        "name": "_gold_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        return cols

    def read_source(self, spark) -> DataFrame:
        """
        Read data source specified in `self.source`

        Parameters
        ----------
        spark: SparkSession
            Spark session

        Returns
        -------
        :
            Output dataframe
        """
        return self.source.read(spark)

    def build_columns(
        self, df: DataFrame, udfs: list[Callable] = None, raise_exception: bool = True
    ) -> DataFrame:
        """
        Build dataframe columns

        Parameters
        ----------
        df:
            Input DataFrame
        udfs:
            User-defined functions
        raise_exception
            If `True`, raise exception when input columns are not available,
            else, skip.
        """
        logger.info(f"Setting columns...")
        built_cols = []
        for col in self._columns_to_build:
            c = col.to_spark(df, udfs=udfs, raise_exception=raise_exception)
            if c is not None:
                df = df.withColumn(col.name, c)
                built_cols += [col]

        for c in built_cols:
            self._columns_to_build.remove(c)

        return df

    def process(self, df, udfs=None, spark=None) -> DataFrame:
        """
        Build table from source DataFrame by applying joins, aggregations and
        creating new columns.

        Parameters
        ----------
        df:
            Input DataFrame
        udfs:
            User-defined functions
        spark: SparkSession
            Spark sessions.

        Returns
        -------
        :
            output Spark DataFrame
        """
        import pyspark.sql.functions as F

        logger.info(f"Applying {self.layer} transformations")

        # Build columns
        self._columns_to_build = self.columns + self._get_layer_columns(
            layer=self.layer, df=df
        )
        column_names = [c.name for c in self._columns_to_build]
        df = self.build_columns(
            df, udfs=udfs, raise_exception=not (self.has_joins or self.aggregation)
        )

        # Make joins
        for i, join in enumerate(self.joins):
            if i == 0:
                name = self.source.name
            else:
                name = "previous_join"
            join.left = TableDataSource(name=name)
            join.left._df = df
            df = join.execute(spark)

            # Build remaining columns again (in case inputs are found in joins)
            df = self.build_columns(
                df, udfs=udfs, raise_exception=i == len(self.joins) - 1
            )

        # Window filtering
        if self.window_filter:
            df = self.window_filter.execute(df)

        # Drop source columns
        if self.drop_source_columns:
            logger.info(f"Dropping source columns...")
            df = df.select(column_names)

        if self.aggregation:
            df = self.aggregation.execute(df, udfs=udfs)
            self._columns_to_build += self._get_layer_columns(layer=self.layer, df=df)

        # Build columns after aggregation
        df = self.build_columns(
            df, udfs=udfs, raise_exception=not self.has_joins_post_aggregation
        )

        # Make post-aggregation joins
        for i, join in enumerate(self.joins_post_aggregation):
            if i == 0:
                name = self.source.name
            else:
                name = "previous_join"
            join.left = TableDataSource(name=name)
            join.left._df = df
            df = join.execute(spark)

            # Build remaining columns again (in case inputs are found in joins)
            df = self.build_columns(
                df, udfs=udfs, raise_exception=i == len(self.joins) - 1
            )

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Select columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Drop columns
        if self.drop_columns:
            logger.info(f"Dropping columns {self.drop_columns}...")
            df = df.drop(*self.drop_columns)

        # Drop duplicates
        pk = self.primary_key
        if self.drop_duplicates and pk:
            logger.info(f"Removing duplicates with {pk}")
            df = df.dropDuplicates([pk])

        return df

    @property
    def apply_changes_kwargs(self) -> dict[str, str]:
        """Keyword arguments for dlt.apply_changes function"""
        cdc = self.source.cdc
        return {
            "apply_as_deletes": cdc.apply_as_deletes,
            "apply_as_truncates": cdc.apply_as_truncates,
            "column_list": cdc.columns,
            "except_column_list": cdc.except_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.sequence_by,
            "source": self.source.name,
            "stored_as_scd_type": cdc.scd_type,
            "target": self._table.name,
            "track_history_column_list": cdc.track_history_columns,
            "track_history_except_column_list": cdc.track_history_except_columns,
        }



from pydantic import ValidationError
from pyspark.sql import types as T
from pyspark.sql import SparkSession
import pandas as pd
from pandas import Timestamp
import pytest

from laktory._testing import EventsManager
from laktory._testing import df_meta
from laktory._testing import table_brz
from laktory._testing import table_slv
from laktory._testing import table_slv_star
from laktory._testing import table_gld
from laktory.models import Table
from laktory.models import TableJoin
from laktory.models import TableAggregation

manager = EventsManager()
manager.build_events()

# Spark
spark = SparkSession.builder.appName("UnitTesting").getOrCreate()


def test_model():
    print(table_slv.model_dump())
    assert table_slv.model_dump() == {
        "builder": {
            "aggregation": None,
            "drop_columns": [],
            "drop_duplicates": None,
            "drop_source_columns": True,
            "event_source": None,
            "filter": None,
            "joins": [],
            "joins_post_aggregation": [],
            "layer": "SILVER",
            "pipeline_name": None,
            "selects": None,
            "table_source": {
                "read_as_stream": True,
                "catalog_name": "dev",
                "cdc": None,
                "selects": None,
                "filter": None,
                "from_pipeline": True,
                "name": "brz_stock_prices",
                "schema_name": "markets",
                "watermark": None,
            },
            "template": "SILVER",
            "window_filter": None,
        },
        "catalog_name": "dev",
        "columns": [
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "created_at",
                "pii": None,
                "schema_name": "markets",
                "spark_func_args": [
                    {
                        "value": "data._created_at",
                        "is_column": True,
                        "to_lit": False,
                        "to_expr": True,
                    }
                ],
                "spark_func_kwargs": {},
                "spark_func_name": "coalesce",
                "sql_expression": None,
                "table_name": "slv_stock_prices",
                "type": "timestamp",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "symbol",
                "pii": None,
                "schema_name": "markets",
                "spark_func_args": [
                    {
                        "value": "data.symbol",
                        "is_column": True,
                        "to_lit": False,
                        "to_expr": True,
                    }
                ],
                "spark_func_kwargs": {},
                "spark_func_name": "coalesce",
                "sql_expression": None,
                "table_name": "slv_stock_prices",
                "type": "string",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "open",
                "pii": None,
                "schema_name": "markets",
                "spark_func_args": [
                    {
                        "value": "data.open",
                        "is_column": True,
                        "to_lit": False,
                        "to_expr": True,
                    }
                ],
                "spark_func_kwargs": {},
                "spark_func_name": "coalesce",
                "sql_expression": None,
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "close",
                "pii": None,
                "schema_name": "markets",
                "spark_func_args": [],
                "spark_func_kwargs": {},
                "spark_func_name": None,
                "sql_expression": "data.open",
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
        ],
        "comment": None,
        "data": [
            ["2023-11-01T00:00:00Z", "AAPL", 1, 2],
            ["2023-11-01T01:00:00Z", "AAPL", 3, 4],
            ["2023-11-01T00:00:00Z", "GOOGL", 3, 4],
            ["2023-11-01T01:00:00Z", "GOOGL", 5, 6],
        ],
        "data_source_format": "DELTA",
        "expectations": [
            {"name": "positive_price", "expression": "open > 0", "action": "FAIL"},
            {
                "name": "recent_price",
                "expression": "created_at > '2023-01-01'",
                "action": "DROP",
            },
        ],
        "grants": None,
        "name": "slv_stock_prices",
        "primary_key": None,
        "schema_name": "markets",
        "table_type": "MANAGED",
        "timestamp_key": None,
        "view_definition": None,
        "warehouse_id": "08b717ce051a0261",
    }

    assert not table_slv.is_from_cdc

    assert table_slv.warning_expectations == {}
    assert table_slv.drop_expectations == {"recent_price": "created_at > '2023-01-01'"}
    assert table_slv.fail_expectations == {"positive_price": "open > 0"}

    # Invalid layer
    with pytest.raises(ValidationError):
        Table(name="googl", layer="ROUGE")


def test_data():
    data = table_slv.to_df().to_dict(orient="records")
    assert data == [
        {"created_at": "2023-11-01T00:00:00Z", "symbol": "AAPL", "open": 1, "close": 2},
        {"created_at": "2023-11-01T01:00:00Z", "symbol": "AAPL", "open": 3, "close": 4},
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
        },
        {
            "created_at": "2023-11-01T01:00:00Z",
            "symbol": "GOOGL",
            "open": 5,
            "close": 6,
        },
    ]


def test_bronze():
    df0 = manager.to_spark_df()
    df1 = table_brz.builder.process(df0)
    assert "_bronze_at" in df1.columns


def test_silver():
    df0 = manager.to_spark_df()
    df0.show()
    df1 = table_brz.builder.process(df0)
    df1.show()
    df2 = table_slv.builder.process(df1)
    df2.show()
    assert df2.schema == T.StructType(
        [
            T.StructField("created_at", T.TimestampType(), True),
            T.StructField("symbol", T.StringType(), True),
            T.StructField("open", T.DoubleType(), True),
            T.StructField("close", T.DoubleType(), True),
            T.StructField("_bronze_at", T.TimestampType(), False),
            T.StructField("_silver_at", T.TimestampType(), False),
        ]
    )
    s = df2.toPandas().iloc[0]
    print(s)
    assert s["created_at"] == pd.Timestamp("2023-09-01 00:00:00")
    assert s["symbol"] == "AAPL"
    assert s["open"] == 189.49000549316406
    assert s["close"] == 189.49000549316406


def test_table_join():
    join = TableJoin(
        left={
            "name": "slv_stock_prices",
            "filter": "created_at = '2023-11-01T00:00:00Z'",
        },
        other={
            "name": "slv_stock_metadata",
            "selects": {
                "symbol2": "symbol",
                "currency": "currency",
                "first_traded": "first_traded",
            },
        },
        on=["symbol"],
    )
    join.left._df = table_slv.to_df(spark)
    join.other._df = df_meta

    df = join.execute(spark)
    data = df.toPandas().to_dict(orient="records")
    print(data)
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "AAPL",
            "open": 1,
            "close": 2,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
            "currency": "USD",
            "first_traded": "2004-08-19T13:30:00.00Z",
        },
    ]


def test_table_agg():
    agg = TableAggregation(
        groupby_columns=["symbol"],
        agg_expressions=[
            {"name": "min_open", "spark_func_name": "min", "spark_func_args": ["open"]},
            {
                "name": "max_open",
                "sql_expression": "max(open)",
            },
        ],
    )
    df = table_slv.to_df(spark)

    df1 = agg.execute(df=df)
    df1.show()
    data = df1.toPandas().to_dict(orient="records")
    print(data)
    assert data == [
        {"symbol": "AAPL", "min_open": 1, "max_open": 3},
        {"symbol": "GOOGL", "min_open": 3, "max_open": 5},
    ]


def test_silver_star():
    print(table_slv_star.model_dump())
    assert table_slv_star.model_dump() == {
        "builder": {
            "aggregation": None,
            "drop_columns": [],
            "drop_duplicates": None,
            "drop_source_columns": False,
            "event_source": None,
            "filter": None,
            "joins": [
                {
                    "how": "left",
                    "left": None,
                    "on": ["symbol"],
                    "other": {
                        "read_as_stream": True,
                        "catalog_name": "dev",
                        "cdc": None,
                        "selects": {
                            "symbol2": "symbol",
                            "currency": "currency",
                            "first_traded": "last_traded",
                        },
                        "filter": None,
                        "from_pipeline": True,
                        "name": "slv_stock_metadata",
                        "schema_name": "markets",
                        "watermark": None,
                    },
                    "time_constraint_interval_lower": "60 seconds",
                    "time_constraint_interval_upper": None,
                },
                {
                    "how": "left",
                    "left": None,
                    "on": ["symbol3"],
                    "other": {
                        "read_as_stream": True,
                        "catalog_name": "dev",
                        "cdc": None,
                        "selects": ["symbol3", "name"],
                        "filter": None,
                        "from_pipeline": True,
                        "name": "slv_stock_names",
                        "schema_name": "markets",
                        "watermark": None,
                    },
                    "time_constraint_interval_lower": "60 seconds",
                    "time_constraint_interval_upper": None,
                },
            ],
            "joins_post_aggregation": [],
            "layer": "SILVER_STAR",
            "pipeline_name": None,
            "selects": None,
            "table_source": {
                "read_as_stream": True,
                "catalog_name": "dev",
                "cdc": None,
                "selects": None,
                "filter": "created_at = '2023-11-01T00:00:00Z'",
                "from_pipeline": True,
                "name": "slv_stock_prices",
                "schema_name": "markets",
                "watermark": None,
            },
            "template": "SILVER_STAR",
            "window_filter": None,
        },
        "catalog_name": "dev",
        "columns": [
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "symbol3",
                "pii": None,
                "schema_name": "markets",
                "spark_func_args": [
                    {
                        "value": "symbol",
                        "is_column": True,
                        "to_lit": False,
                        "to_expr": True,
                    }
                ],
                "spark_func_kwargs": {},
                "spark_func_name": "coalesce",
                "sql_expression": None,
                "table_name": "slv_star_stock_prices",
                "type": "string",
                "unit": None,
            }
        ],
        "comment": None,
        "data": None,
        "data_source_format": "DELTA",
        "expectations": [],
        "grants": None,
        "name": "slv_star_stock_prices",
        "primary_key": None,
        "schema_name": "markets",
        "table_type": "MANAGED",
        "timestamp_key": None,
        "view_definition": None,
        "warehouse_id": "08b717ce051a0261",
    }

    df = table_slv_star.builder.read_source(spark)
    df = table_slv_star.builder.process(df, spark=spark)
    assert "_silver_star_at" in df.columns
    data = df.toPandas().drop("_silver_star_at", axis=1).to_dict("records")
    print(data)
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "AAPL",
            "open": 1,
            "close": 2,
            "symbol3": "AAPL",
            "currency": "USD",
            "last_traded": "1980-12-12T14:30:00.000Z",
            "name": "Apple",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
            "symbol3": "GOOGL",
            "currency": "USD",
            "last_traded": "2004-08-19T13:30:00.00Z",
            "name": "Google",
        },
    ]


def test_gold():
    assert table_gld.builder.template == "GOLD1"
    df0 = manager.to_spark_df()
    df0.show()
    df1 = table_brz.builder.process(df0)
    df1.show()
    df2 = table_slv.builder.process(df1)
    df2.show()
    df3 = table_gld.builder.process(df2)
    df3.show()
    df3.printSchema()
    assert df3.schema == T.StructType(
        [
            T.StructField("min_open", T.DoubleType(), True),
            T.StructField("max_open", T.DoubleType(), True),
            T.StructField("min_close", T.DoubleType(), True),
            T.StructField("_gold_at", T.TimestampType(), False),
            T.StructField("name", T.StringType(), True),
            T.StructField("symbol", T.StringType(), True),
            T.StructField("name2", T.StringType(), True),
        ]
    )
    s = df3.toPandas().iloc[0]
    print(s)
    assert s["name2"] == "Apple"


def test_cdc():
    table = Table(
        name="brz_users_type1",
        builder={
            "table_source": {
                "name": "brz_users_cdc",
                "cdc": {
                    "primary_keys": ["userId"],
                    "sequence_by": "sequenceNum",
                    "apply_as_deletes": "operation = 'DELETE'",
                    "scd_type": 1,
                    "except_columns": ["operation", "sequenceNum"],
                },
            },
        },
    )

    assert table.builder.apply_changes_kwargs == {
        "apply_as_deletes": "operation = 'DELETE'",
        "apply_as_truncates": None,
        "column_list": [],
        "except_column_list": ["operation", "sequenceNum"],
        "ignore_null_updates": None,
        "keys": ["userId"],
        "sequence_by": "sequenceNum",
        "source": "brz_users_cdc",
        "stored_as_scd_type": 1,
        "target": "brz_users_type1",
        "track_history_column_list": None,
        "track_history_except_column_list": None,
    }
    assert table.builder.is_from_cdc

    # TODO: Run test with demo data
    # from pyspark.sql import SparkSession

    # spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
    #
    # df_cdc = spark.createDataFrame(pd.DataFrame({
    #     "userId": [124, 123, 125, 126, 123, 125, 125, 123],
    #     "name": ["Raul", "Isabel", "Mercedes", "Lily", None, "Mercedes", "Mercedes", "Isabel"],
    #     "city": ["Oaxaca", "Monterrey", "Tijuana", "Cancun", None, "Guadalajara", "Mexicali", "Chihuahua"],
    #     "operation": ["INSERT", "INSERT", "INSERT", "INSERT", "DELETE", "UPDATE", "UPDATE", "UPDATE"],
    #     "sequenceNum": [1, 1, 2, 2, 6, 6, 5, 5],
    # }))
    # df_cdc.show()


if __name__ == "__main__":
    test_model()
    test_data()
    test_bronze()
    test_silver()
    test_table_join()
    test_table_agg()
    test_silver_star()
    test_gold()
    test_cdc()



import pydantic


class Model(pydantic.BaseModel):
    a: int

    # @pydantic.computed_field(repr=False)
    @property
    def b(self) -> int:
        return self.a + 1


m = Model(a=5)

print(m)
print(m.model_dump())



import laktory
from laktory.models import BaseModel


class Camel(BaseModel):
    d: dict = {}
    l: list = []


def test_camel_case():
    camel = Camel(
        d={"this_is_a_test": ["value_a", "value_b", {"key_alpha": 0, "keyBeta": 1}]},
        l=["a", "a-b-c", "class_member"],
    )
    d = laktory._parsers.camelize_keys(camel.model_dump())
    print(d)
    assert d == {
        "d": {"thisIsATest": ["value_a", "value_b", {"keyBeta": 1, "keyAlpha": 0}]},
        "l": ["a", "a-b-c", "class_member"],
    }


if __name__ == "__main__":
    test_camel_case()



import copy


def _snake_to_camel(snake_str):
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def camelize_keys(d, parent=None, excluded_parents=None):
    if isinstance(d, dict):
        if parent and excluded_parents and parent in excluded_parents:
            return d
        keys = list(d.keys())
        values = list(d.values())
        for key, value in zip(keys, values):
            new_key = _snake_to_camel(key)
            d[new_key] = camelize_keys(
                value, parent=key, excluded_parents=excluded_parents
            )
            if new_key != key:
                del d[key]

    elif isinstance(d, list):
        for i, item in enumerate(d):
            d[i] = camelize_keys(item)
    else:
        pass
    return d


def merge_dicts(d1: dict, d2: dict) -> dict:
    dm = copy.deepcopy(d1)

    def _merge_dicts(d1, d2):
        for key, value in d2.items():
            if key in d1 and isinstance(d1[key], dict) and isinstance(value, dict):
                _merge_dicts(d1[key], value)
            else:
                d1[key] = value

    _merge_dicts(dm, d2)

    return dm



import re
from pyspark.sql.dataframe import DataFrame

from laktory.spark.dataframe.schema_flat import schema_flat


def has_column(df: DataFrame, col: str) -> bool:
    """
    Check if column `col` exists in `df`

    Parameters
    ----------
    df:
        Input DataFrame
    col
        Column name

    Returns
    -------
    :
        Result

    Examples
    --------

    ```py
    import laktory  # noqa: F401
    import pyspark.sql.types as T

    schema = T.StructType(
        [
            T.StructField("indexx", T.IntegerType()),
            T.StructField(
                "stock",
                T.StructType(
                    [
                        T.StructField("symbol", T.StringType()),
                        T.StructField("name", T.StringType()),
                    ]
                ),
            ),
            T.StructField(
                "prices",
                T.ArrayType(
                    T.StructType(
                        [
                            T.StructField("open", T.IntegerType()),
                            T.StructField("close", T.IntegerType()),
                        ]
                    )
                ),
            ),
        ]
    )

    data = [
        (
            1,
            {"symbol": "AAPL", "name": "Apple"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
        (
            2,
            {"symbol": "MSFT", "name": "Microsoft"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
        (
            3,
            {"symbol": "GOOGL", "name": "Google"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
    ]

    df = spark.createDataFrame(data, schema=schema)
    print(df.has_column("symbol"))
    #> False
    print(df.has_column("`stock`.`symbol`"))
    #> True
    print(df.has_column("`prices[2]`.`close`"))
    #> True
    ```
    """
    _col = re.sub(r"\[(\d+)\]", r"[*]", col)
    _col = re.sub(r"`", "", _col)
    return _col in schema_flat(df)



import os

from laktory.models import DatabricksProvider


def test_databricks_provider():
    p = DatabricksProvider(host="databricks.net")
    print(p)

    print(p.resource_key)
    print(p.resource_name)


if __name__ == "__main__":
    test_databricks_provider()



from .catalog import Catalog
from .column import Column
from .schema import Schema
from .table import Table
from .tableaggregation import TableAggregation
from .tablebuilder import TableBuilder
from .tableexpectation import TableExpectation
from .tablejoin import TableJoin
from .tablewindowfilter import TableWindowFilter
from .volume import Volume



from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.cluster import Cluster
from laktory.models.databricks.directory import Directory
from laktory.models.databricks.grants import Grants
from laktory.models.databricks.group import Group
from laktory.models.databricks.job import Job
from laktory.models.databricks.metastore import Metastore
from laktory.models.databricks.metastoreassignment import MetastoreAssignment
from laktory.models.databricks.mwspermissionassignment import MwsPermissionAssignment
from laktory.models.databricks.notebook import Notebook
from laktory.models.databricks.pipeline import Pipeline
from laktory.models.databricks.secret import Secret
from laktory.models.databricks.secretacl import SecretAcl
from laktory.models.databricks.secretscope import SecretScope
from laktory.models.databricks.serviceprincipal import ServicePrincipal
from laktory.models.databricks.serviceprincipalrole import ServicePrincipalRole
from laktory.models.databricks.sqlquery import SqlQuery
from laktory.models.databricks.user import User
from laktory.models.databricks.userrole import UserRole
from laktory.models.databricks.warehouse import Warehouse
from laktory.models.databricks.workspacefile import WorkspaceFile



import json
from pyspark.sql.dataframe import DataFrame


def schema_flat(df: DataFrame) -> list[str]:
    """
    Returns a flattened list of columns

    Parameters
    ----------
    df:
        Input DataFrame

    Returns
    -------
    :
        List of columns

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import pyspark.sql.types as T

    schema = T.StructType(
        [
            T.StructField("indexx", T.IntegerType()),
            T.StructField(
                "stock",
                T.StructType(
                    [
                        T.StructField("symbol", T.StringType()),
                        T.StructField("name", T.StringType()),
                    ]
                ),
            ),
            T.StructField(
                "prices",
                T.ArrayType(
                    T.StructType(
                        [
                            T.StructField("open", T.IntegerType()),
                            T.StructField("close", T.IntegerType()),
                        ]
                    )
                ),
            ),
        ]
    )

    data = [
        (
            1,
            {"symbol": "AAPL", "name": "Apple"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
        (
            2,
            {"symbol": "MSFT", "name": "Microsoft"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
        (
            3,
            {"symbol": "GOOGL", "name": "Google"},
            [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
        ),
    ]

    df = spark.createDataFrame(data, schema=schema)
    print(df.schema_flat())
    '''
    [
        'indexx',
        'stock',
        'stock.symbol',
        'stock.name',
        'prices',
        'prices[*].open',
        'prices[*].close',
    ]
    '''
    ```
    """

    def get_fields(json_schema):
        field_names = []
        for f in json_schema.get("fields", []):
            f_name = f["name"]
            f_type = f["type"]
            if isinstance(f_type, dict):
                if f_type["type"] == "array":
                    e_type = f_type["elementType"]
                    field_names += [f_name]
                    if isinstance(e_type, dict):
                        _field_names = get_fields(f_type["elementType"])
                        field_names += [f"{f_name}[*].{v}" for v in _field_names]
                elif f_type["type"] == "struct":
                    _field_names = get_fields(f["type"])
                    field_names += [f_name]
                    field_names += [f"{f_name}.{v}" for v in _field_names]
                elif f_type["type"] == "map":
                    field_names += [f_name]
                else:
                    raise ValueError(f_type["type"])
            else:
                field_names += [f_name]
        return field_names

    return get_fields(json.loads(df.schema.json()))



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Directory(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Directory

    Attributes
    ----------
    path:
        The absolute path of the directory, beginning with "/", e.g. "/pipelines".
    delete_recursive:
        When `True`, subdirectories are also deleted when the directory is deleted

    Examples
    --------
    ```py
    from laktory import models

    d = models.Directory(path="/queries/views")
    print(d)
    '''
    resource_name_=None options=ResourceOptions(variables={}, depends_on=[], provider=None, aliases=None, delete_before_replace=True, ignore_changes=None, import_=None, parent=None, replace_on_changes=None) variables={} path='/queries/views' delete_recursive=None
    '''
    print(d.resource_key)
    #> queries-views
    print(d.resource_name)
    #> directory-queries-views
    ```
    """

    path: str
    delete_recursive: Union[bool, None] = None

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        key = self.path
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        for i in range(5):
            if key.endswith("-"):
                key = key[:-1]
        return key

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Directory"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Directory

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_directory"



import os
from typing import Any
from typing import Union
from pydantic import model_validator
from laktory import constants
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.permissions import Permissions


class WorkspaceFile(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Workspace File

    Attributes
    ----------
    access_controls:
        List of file access controls
    dirpath:
        Workspace directory containing the file. Filename will be assumed to be the same as local filepath. Used if path
        is not specified.
    path:
         Workspace filepath for the file
    source:
        Path to file on local filesystem.
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    path: str = None
    source: str

    @property
    def filename(self) -> str:
        """File filename"""
        return os.path.basename(self.source)

    @model_validator(mode="after")
    def default_path(self) -> Any:
        if self.path is None:
            if self.dirpath:
                self.path = f"{self.dirpath}{self.filename}"

            elif "/workspacefiles/" in self.source:
                self.path = (
                    constants.LAKTORY_WORKSPACE_ROOT
                    + self.source.split("/workspacefiles/")[-1]
                )

            else:
                raise ValueError(
                    "A value for `dirpath` must be specified if the source is not in a `workspacefiles` folder"
                )

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        # key = os.path.splitext(self.path)[0]
        key = self.path
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    workspace_file_path=self.path,
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:WorkspaceFile"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.WorkspaceFile

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_workspace_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from abc import abstractmethod
from typing import Union
from laktory.spark import DataFrame

from laktory.models.basemodel import BaseModel


class BaseDataSource(BaseModel):
    """
    Base class for building data source

    Attributes
    ----------
    read_as_stream
        If `True` read source as stream
    """

    read_as_stream: Union[bool, None] = True

    @abstractmethod
    def read(self, spark) -> DataFrame:
        raise NotImplementedError()

    @property
    def is_cdc(self) -> bool:
        """If `True` source data is a change data capture (CDC)"""
        return getattr(self, "cdc", None) is not None



from typing import Union

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class OrderBy(BaseModel):
    sql_expression: str
    desc: bool = False


class TableWindowFilter(BaseModel):
    """
    Specifications of a window-based filtering.

    Attributes
    ----------
    descending:
        If `True` rows are sorted in descending order before dropping rows
        from each window.
    drop_row_index:
        If `True`, the row index column is dropped
    order_by:
        Defines the column used for sorting before dropping rows
    partition_by
        Defines the columns used for grouping into Windows
    row_index_name:
        Group-specific and sorted rows index name
    rows_to_keep:
        How many rows to keep per window

    Examples
    --------
    ```py
    from datetime import datetime
    from laktory import models

    df0 = spark.createDataFrame(
        [
            [datetime(2023, 1, 1), "APPL", 200.0],
            [datetime(2023, 1, 2), "APPL", 202.0],
            [datetime(2023, 1, 3), "APPL", 201.5],
            [datetime(2023, 1, 1), "GOOL", 200.0],
            [datetime(2023, 1, 2), "GOOL", 202.0],
            [datetime(2023, 1, 3), "GOOL", 201.5],
        ],
        ["created_at", "symbol", "price"],
    )

    f = models.TableWindowFilter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )
    df1 = f.execute(df0)
    ```

    References
    ----------

    * [pyspark Window](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Window.html)
    """

    descending: bool = True
    drop_row_index: bool = True
    order_by: Union[list[OrderBy], None] = None
    partition_by: Union[list[str], None]
    row_index_name: str = "_row_index"
    rows_to_keep: int = 1

    def execute(self, df):
        """
        Execute window filtering

        Parameters
        ----------
        df: DataFrame
            Spark DataFrame

        Returns
        -------
        : DataFrame
            Output DataFrame
        """
        from pyspark.sql import Window
        import pyspark.sql.functions as F

        # Partition by
        w = Window.partitionBy(*self.partition_by)

        # Order by
        if self.order_by:
            order_bys = []
            for o in self.order_by:
                order_bys += [F.expr(o.sql_expression)]
                if o.desc:
                    order_bys[-1] = order_bys[-1].desc()
            w = w.orderBy(*order_bys)

        # Set rows index
        df = df.withColumn(self.row_index_name, F.row_number().over(w))

        # Filter
        df = df.filter(F.col(self.row_index_name) <= self.rows_to_keep)
        if self.drop_row_index:
            df = df.drop(self.row_index_name)

        return df


if __name__ == "__main__":
    from laktory import models

    f = models.TableWindowFilter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )
    df1 = f.execute(df0)



from laktory.models import Notebook

nb = Notebook(
    source="./test_notebook.py",
    path="/laktory/hello",
)


def test_notebook():

    nb0 = Notebook(source="./hello_world.py")
    nb1 = Notebook(source="./notebooks/demos/hello_world.py")
    nb2 = Notebook(source="./notebooks/demos/hello_world.py", dirpath="/files/")

    assert nb.path == "/laktory/hello"
    assert nb.source == "./test_notebook.py"
    assert nb0.path is None
    assert nb1.path == "/.laktory/demos/hello_world.py"
    assert nb2.path == "/files/hello_world.py"
    assert nb1.resource_key == "laktory-demos-hello_world-py"
    assert nb1.resource_name == "notebook-laktory-demos-hello_world-py"


if __name__ == "__main__":
    test_notebook()



import os

from laktory.models import Job
from laktory import pulumi_outputs


root_dir = os.path.dirname(__file__)

job = Job(
    name="job-stock-prices",
    clusters=[
        {
            "name": "main",
            "spark_version": "14.0.x-scala2.12",
            "node_type_id": "Standard_DS3_v2",
        },
    ],
    tasks=[
        {
            "job_cluster_key": "main",
            "notebook_task": {
                "notebook_path": "job/ingest_stock_prices",
            },
            "task_key": "ingestion",
        },
        {
            "depends_ons": [{"task_key": "ingestion"}],
            "pipeline_task": {"pipeline_id": "${resources.pl-stock-prices.id}"},
            "task_key": "pipeline",
        },
        {
            "task_key": "view",
            "sql_task": {
                "query": {"query_id": "456"},
                "warehouse_id": "123",
            },
        },
    ],
)


def test_job_model():
    data = job.model_dump()
    print(data)
    assert data == {
        "access_controls": [],
        "clusters": [
            {
                "apply_policy_default_values": None,
                "autoscale": None,
                "autotermination_minutes": None,
                "cluster_id": None,
                "custom_tags": None,
                "data_security_mode": "USER_ISOLATION",
                "driver_instance_pool_id": None,
                "driver_node_type_id": None,
                "enable_elastic_disk": None,
                "enable_local_disk_encryption": None,
                "idempotency_token": None,
                "init_scripts": [],
                "instance_pool_id": None,
                "name": "main",
                "node_type_id": "Standard_DS3_v2",
                "num_workers": None,
                "policy_id": None,
                "runtime_engine": None,
                "single_user_name": None,
                "spark_conf": {},
                "spark_env_vars": {},
                "spark_version": "14.0.x-scala2.12",
                "ssh_public_keys": [],
            }
        ],
        "continuous": None,
        "control_run_state": None,
        "email_notifications": None,
        "format": None,
        "health": None,
        "max_concurrent_runs": None,
        "max_retries": None,
        "min_retry_interval_millis": None,
        "name": "job-stock-prices",
        "notification_settings": None,
        "parameters": [],
        "retry_on_timeout": None,
        "run_as": None,
        "schedule": None,
        "tags": {},
        "tasks": [
            {
                "condition_task": None,
                "depends_ons": None,
                "description": None,
                "email_notifications": None,
                "existing_cluster_id": None,
                "health": None,
                "job_cluster_key": "main",
                "libraries": None,
                "max_retries": None,
                "min_retry_interval_millis": None,
                "notebook_task": {
                    "notebook_path": "job/ingest_stock_prices",
                    "base_parameters": None,
                    "source": None,
                },
                "notification_settings": None,
                "pipeline_task": None,
                "retry_on_timeout": None,
                "run_if": None,
                "run_job_task": None,
                "sql_task": None,
                "task_key": "ingestion",
                "timeout_seconds": None,
            },
            {
                "condition_task": None,
                "depends_ons": [{"task_key": "ingestion", "outcome": None}],
                "description": None,
                "email_notifications": None,
                "existing_cluster_id": None,
                "health": None,
                "job_cluster_key": None,
                "libraries": None,
                "max_retries": None,
                "min_retry_interval_millis": None,
                "notebook_task": None,
                "notification_settings": None,
                "pipeline_task": {
                    "pipeline_id": "${resources.pl-stock-prices.id}",
                    "full_refresh": None,
                },
                "retry_on_timeout": None,
                "run_if": None,
                "run_job_task": None,
                "sql_task": None,
                "task_key": "pipeline",
                "timeout_seconds": None,
            },
            {
                "condition_task": None,
                "depends_ons": None,
                "description": None,
                "email_notifications": None,
                "existing_cluster_id": None,
                "health": None,
                "job_cluster_key": None,
                "libraries": None,
                "max_retries": None,
                "min_retry_interval_millis": None,
                "notebook_task": None,
                "notification_settings": None,
                "pipeline_task": None,
                "retry_on_timeout": None,
                "run_if": None,
                "run_job_task": None,
                "sql_task": {
                    "alert": None,
                    "dashboard": None,
                    "file": None,
                    "parameters": None,
                    "query": {"query_id": "456"},
                    "warehouse_id": "123",
                },
                "task_key": "view",
                "timeout_seconds": None,
            },
        ],
        "timeout_seconds": None,
        "trigger": None,
        "webhook_notifications": None,
    }


def test_job_pulumi():
    pulumi_outputs["pl-stock-prices.id"] = "12345"
    assert job.resource_name == "job-stock-prices"
    assert job.options.model_dump(exclude_none=True) == {
        "depends_on": [],
        "delete_before_replace": True,
    }
    data = job.pulumi_properties
    print(data)
    assert data == {
        "name": "job-stock-prices",
        "parameters": [],
        "tags": {},
        "tasks": [
            {
                "job_cluster_key": "main",
                "notebook_task": {"notebook_path": "job/ingest_stock_prices"},
                "task_key": "ingestion",
            },
            {
                "depends_ons": [{"task_key": "ingestion"}],
                "pipeline_task": {"pipeline_id": "12345"},
                "task_key": "pipeline",
            },
            {
                "sql_task": {"query": {"query_id": "456"}, "warehouse_id": "123"},
                "task_key": "view",
            },
        ],
        "job_clusters": [
            {
                "job_cluster_key": "main",
                "new_cluster": {
                    "data_security_mode": "USER_ISOLATION",
                    "init_scripts": [],
                    "node_type_id": "Standard_DS3_v2",
                    "spark_conf": {},
                    "spark_env_vars": {},
                    "spark_version": "14.0.x-scala2.12",
                    "ssh_public_keys": [],
                },
            }
        ],
    }


if __name__ == "__main__":
    test_job_model()
    test_job_pulumi()



import json
import os
from typing import Any
from typing import Literal
from typing import Union
from pydantic import model_validator
from pydantic import Field

from laktory._settings import settings
from laktory.constants import CACHE_ROOT
from laktory.models.basemodel import BaseModel
from laktory.models.databricks.cluster import Cluster
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.workspacefile import WorkspaceFile
from laktory.models.sql.table import Table


class PipelineLibraryFile(BaseModel):
    """
    Pipeline Library File specifications

    Attributes
    ----------
    path:
        Workspace filepath
    """

    path: str


class PipelineLibraryNotebook(BaseModel):
    """
    Pipeline Library Notebook specifications

    Attributes
    ----------
    path:
        Workspace notebook filepath
    """

    path: str


class PipelineLibrary(BaseModel):
    """
    Pipeline Library specifications

    Attributes
    ----------
    file:
        File specifications
    notebook:
        Notebook specifications
    """

    file: str = None
    notebook: PipelineLibraryNotebook = None


class PipelineNotifications(BaseModel):
    """
    Pipeline Notifications specifications

    Attributes
    ----------
    alerts:
        Alert types
    recipients:
        List of user/group/service principal names
    """

    alerts: list[
        Literal[
            "on-update-success",
            "on-update-failure",
            "on-update-fatal-failure",
            "on-flow-failure",
        ]
    ]
    recipients: list[str]


class PipelineCluster(Cluster):
    """
    Pipeline Cluster. Same attributes as `laktory.models.Cluster`, except for

    * `autotermination_minutes`
    * `cluster_id`
    * `data_security_mode`
    * `enable_elastic_disk`
    * `idempotency_token`
    * `is_pinned`
    * `libraries`
    * `node_type_id`
    * `runtime_engine`
    * `single_user_name`
    * `spark_version`

    that are not allowed.
    """

    autotermination_minutes: int = Field(None, exclude=True)
    cluster_id: str = Field(None, exclude=True)
    data_security_mode: str = Field(None, exclude=True)
    enable_elastic_disk: bool = Field(None, exclude=True)
    idempotency_token: str = Field(None, exclude=True)
    is_pinned: bool = Field(None, exclude=True)
    libraries: list[Any] = Field(None, exclude=True)
    node_type_id: str = None
    runtime_engine: str = Field(None, exclude=True)
    single_user_name: str = Field(None, exclude=True)
    spark_version: str = Field(None, exclude=True)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "autotermination_minutes",
            "cluster_id",
            "data_security_mode",
            "enable_elastic_disk",
            "idempotency_token",
            "is_pinned",
            "libraries",
            "runtime_engine",
            "single_user_name",
            "spark_version",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class PipelineUDF(BaseModel):
    """
    Pipeline User Define Function

    Attributes
    ----------
    module_name:
        Name of the module from which the function needs to be imported.
    function_name:
        Name of the function.
    module_path:
        Workspace filepath of the module, if not in the same directory as the pipeline notebook
    """

    module_name: str
    function_name: str
    module_path: str = None


class Pipeline(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Delta Live Tables (DLT) Pipeline

    Attributes
    ----------
    access_controls:
        Pipeline access controls
    allow_duplicate_names:
        If `False`, deployment will fail if name conflicts with that of another pipeline.
    catalog:
        Name of the unity catalog storing the pipeline tables
    channel:
        Name of the release channel for Spark version used by DLT pipeline.
    clusters:
        Clusters to run the pipeline. If none is specified, pipelines will automatically select a default cluster
        configuration for the pipeline.
    configuration:
         List of values to apply to the entire pipeline. Elements must be formatted as key:value pairs
    continuous:
        If `True`, the pipeline is run continuously.
    development:
        If `True` the pipeline is run in development mode
    edition:
        Name of the product edition
    libraries:
        Specifies pipeline code (notebooks) and required artifacts.
    name:
        Pipeline name
    notifications:
        Notifications specifications
    photon:
        If `True`, Photon engine enabled.
    serverless:
        If `True`, serverless is enabled
    storage:
        A location on DBFS or cloud storage where output data and metadata required for pipeline execution are stored.
        By default, tables are stored in a subdirectory of this location. Change of this parameter forces recreation
        of the pipeline. (Conflicts with `catalog`).
    tables:
        List of tables to build
    target:
        The name of a database (in either the Hive metastore or in a UC catalog) for persisting pipeline output data.
        Configuring the target setting allows you to view and query the pipeline output data from the Databricks UI.
    udfs:
        List of user defined functions provided to the table builders.

    Examples
    --------
    Assuming the configuration yaml file
    ```py
    import io
    from laktory import models

    # Define pipeline
    pipeline_yaml = '''
    name: pl-stock-prices

    catalog: dev
    target: finance

    clusters:
      - name : default
        node_type_id: Standard_DS3_v2
        autoscale:
          min_workers: 1
          max_workers: 2

    libraries:
      - notebook:
          path: /pipelines/dlt_brz_template.py
      - notebook:
          path: /pipelines/dlt_slv_template.py
      - notebook:
          path: /pipelines/dlt_gld_stock_performances.py

    access_controls:
      - group_name: account users
        permission_level: CAN_VIEW
      - group_name: role-engineers
        permission_level: CAN_RUN

    # --------------------------------------------------------------------------- #
    # Tables                                                                      #
    # --------------------------------------------------------------------------- #

    tables:
      - name: brz_stock_prices
        timestamp_key: data.created_at
        builder:
          layer: BRONZE
          event_source:
            name: stock_price
            producer:
              name: yahoo-finance
            read_as_stream: True

      - name: slv_stock_prices
        timestamp_key: created_at
        builder:
          layer: SILVER
          table_source:
            name: brz_stock_prices
            read_as_stream: True
        columns:
          - name: created_at
            type: timestamp
            spark_func_name: coalesce
            spark_func_args:
              - data._created_at

          - name: symbol
            type: string
            spark_func_name: coalesce
            spark_func_args:
              - data.symbol

          - name: open
            type: double
            spark_func_name: coalesce
            spark_func_args:
              - data.open

          - name: close
            type: double
            spark_func_name: coalesce
            spark_func_args:
              - data.close
    '''
    pipeline = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```

    References
    ----------

    * [Databricks Pipeline](https://docs.databricks.com/api/workspace/pipelines/create)
    * [Pulumi Databricks Pipeline](https://www.pulumi.com/registry/packages/databricks/api-docs/pipeline/)
    """

    access_controls: list[AccessControl] = []
    allow_duplicate_names: bool = None
    catalog: Union[str, None] = None
    channel: Literal["CURRENT", "PREVIEW"] = "PREVIEW"
    clusters: list[PipelineCluster] = []
    configuration: dict[str, str] = {}
    continuous: bool = None
    development: Union[bool, str] = None
    edition: Literal["CORE", "PRO", "ADVANCED"] = None
    # filters
    libraries: list[PipelineLibrary] = None
    name: str
    notifications: list[PipelineNotifications] = []
    photon: bool = None
    serverless: bool = None
    storage: str = None
    tables: list[Table] = []
    target: str = None
    udfs: list[PipelineUDF] = []

    @model_validator(mode="after")
    def assign_pipeline_to_tables(self) -> Any:
        for t in self.tables:
            t.builder.pipeline_name = self.name
            t.catalog_name = self.catalog
            t.schema_name = self.target
            for c in t.columns:
                c.table_name = t.name
                c.catalog_name = t.catalog_name
                c.schema_name = t.schema_name

            # Assign to sources
            if t.builder.table_source is not None:
                if t.builder.table_source.catalog_name is None:
                    t.builder.table_source.catalog_name = self.catalog
                if t.builder.table_source.schema_name is None:
                    t.builder.table_source.schema_name = self.target

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        """
        pl
        """
        return "pl"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        - configuration workspace file
        - configuration workspace file permissions
        """
        resources = []

        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    pipeline_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        # Configuration file
        source = os.path.join(CACHE_ROOT, f"tmp-{self.name}.json")
        d = self.model_dump(exclude_none=True)
        d = self.inject_vars(d)
        s = json.dumps(d, indent=4)
        with open(source, "w", newline="\n") as fp:
            fp.write(s)
        filepath = f"{settings.workspace_laktory_root}pipelines/{self.name}.json"
        file = WorkspaceFile(
            path=filepath,
            source=source,
        )
        resources += [file]

        resources += [
            Permissions(
                resource_name=f"permissions-{file.resource_name}",
                access_controls=[
                    AccessControl(
                        permission_level="CAN_READ",
                        group_name="account users",
                    )
                ],
                workspace_file_path=filepath,
                options={"depends_on": [f"${{resources.{file.resource_name}}}"]},
            )
        ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Pipeline"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Pipeline

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return {
            "access_controls": True,
            "tables": True,
            "clusters": {"__all__": {"access_controls"}},
            "udfs": True,
        }

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        k = "clusters"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("name")
                _clusters += [c]
            d[k] = _clusters
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_pipeline"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        k = "cluster"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("name")
                _clusters += [c]
            d[k] = _clusters
        return d



from typing import Union

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.datasources.tabledatasource import TableDataSource
from laktory.spark import DataFrame

logger = get_logger(__name__)


class TableJoin(BaseModel):
    """
    Specifications of a tables join.

    Attributes
    ----------
    how:
        Type of join (left, outer, full, etc.)
    left:
        Left side of the join
    on:
        A list of strings for the columns to join on. The columns must exist
        on both sides.
    other:
        Right side of the join
    time_constraint_interval_lower:
        Lower bound for a spark streaming event-time constraint
    time_constraint_interval_upper:
        Upper bound for a spark streaming event-time constraint

    Examples
    --------
    ```py
    from laktory import models

    table = models.Table(
        name="slv_star_stock_prices",
        builder={
            "layer": "SILVER",
            "table_source": {
                "name": "slv_stock_prices",
            },
            "joins": [
                {
                    "other": {
                        "name": "slv_stock_metadata",
                        "read_as_stream": False,
                        "selects": ["symbol", "currency", "first_trader"],
                    },
                    "on": ["symbol"],
                }
            ],
        },
    )
    ```

    References
    ----------

    * [pyspark join](https://spark.apache.org/docs/3.1.2/api/python/reference/api/pyspark.sql.DataFrame.join.html)
    * [spark streaming join](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#inner-joins-with-optional-watermarking)
    """

    how: str = "left"
    left: TableDataSource = None
    on: list[str]
    other: TableDataSource
    time_constraint_interval_lower: str = "60 seconds"
    time_constraint_interval_upper: str = None

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def execute(self, spark) -> DataFrame:
        """
        Execute join

        Parameters
        ----------
        spark: SparkSession
            Spark session

        Returns
        -------
        : DataFrame
            Output DataFrame
        """
        import pyspark.sql.functions as F

        logger.info(f"Executing {self.left.name} {self.how} JOIN {self.other.name}")

        left_df = self.left.read(spark)
        other_df = self.other.read(spark)

        # Add watermark
        other_cols = []
        if self.other.watermark is not None:
            other_cols += [F.col(self.other.watermark.column).alias("_other_wc")]
        other_cols += [F.col(c) for c in other_df.columns]
        other_df = other_df.select(other_cols)

        # Drop duplicates to prevent adding rows to left
        other_df = other_df.dropDuplicates(self.on)

        _join = []
        for c in self.on:
            _join += [f"left.{c} == other.{c}"]
        if self.other.watermark is not None:
            if self.time_constraint_interval_lower:
                _join += [
                    f"left.{self.left.watermark.column} >= other._other_wc - interval {self.time_constraint_interval_lower}"
                ]
            if self.time_constraint_interval_upper:
                _join += [
                    f"left.{self.left.watermark.column} <= other._other_wc + interval {self.time_constraint_interval_upper}"
                ]
        _join = " AND ".join(_join)

        logger.info(f"   ON {_join}")

        logger.info(f"Left Schema:")
        left_df.printSchema()

        logger.info(f"Other Schema:")
        other_df.printSchema()

        df = (
            left_df.alias("left")
            .join(
                other=other_df.alias("other"),
                on=F.expr(_join),
                how=self.how,
            )
            .drop()
        )

        # Clean join columns
        for c in self.on:
            df = df.withColumn("__tmp", F.coalesce(f"left.{c}", f"other.{c}"))
            df = df.drop(c)
            df = df.withColumn(c, F.col("__tmp"))
            df = df.drop("__tmp")

        # Drop watermark column
        if self.other.watermark is not None:
            df = df.drop(F.col(f"other._other_wc"))
        logger.info(f"Joined Schema:")
        df.printSchema()

        return df


if __name__ == "__main__":
    from laktory import models

    table = models.Table(
        name="slv_star_stock_prices",
        builder={
            "layer": "SILVER",
            "table_source": {
                "name": "slv_stock_prices",
            },
            "joins": [
                {
                    "other": {
                        "name": "slv_stock_metadata",
                        "read_as_stream": False,
                        "selects": ["symbol", "currency", "first_trader"],
                    },
                    "on": ["symbol"],
                }
            ],
        },
    )

    # Read
    df = table.builder.read_source(spark)

    # Process
    df = table.builder.process(df, spark)



from typing import Literal
from typing import Union
from pydantic import Field
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.permissions import Permissions


class ClusterAutoScale(BaseModel):
    """
    Cluster Autoscale

    Attributes
    ----------
    min_workers:
        Minimum number of worker nodes
    max_workers:
        Maximum number of worker nodes
    """

    min_workers: int
    max_workers: int


class ClusterInitScriptVolumes(BaseModel):
    """
    Cluster Init Script Workspace

    Attributes
    ----------
    destination:
        Volume filepath
    """

    destination: str = None


class ClusterInitScriptWorkspace(BaseModel):
    """
    Cluster Init Script Workspace

    Attributes
    ----------
    destination:
        Workspace filepath
    """

    destination: str = None


class ClusterInitScript(BaseModel):
    """
    Cluster Init Script

    Attributes
    ----------
    volumes:
        Volumes file specification
    workspace:
        Workspace file specifications
    """

    volumes: ClusterInitScriptVolumes = None
    workspace: ClusterInitScriptWorkspace = None


class ClusterLibraryCran(BaseModel):
    package: str = None
    repo: str = None


class ClusterLibraryMaven(BaseModel):
    coordinates: str = None
    exclusions: list[str] = None
    repo: str = None


class ClusterLibraryPypi(BaseModel):
    """
    Cluster Library Pypi

    Attributes
    ----------
    package:
        Package name
    repo:
        Packages repository
    """

    package: str = None
    repo: str = None


class ClusterLibrary(BaseModel):
    """
    Cluster Library

    Attributes
    ----------
    cran:
        Cran library specifications
    egg:
        Egg filepath
    jar:
        Jar filepath
    maven:
    pypi:
        Pypi library specifications
    whl:
        Wheel filepath
    """

    cran: ClusterLibraryCran = None
    egg: str = None
    jar: str = None
    maven: ClusterLibraryMaven = None
    pypi: ClusterLibraryPypi = None
    whl: str = None


class Cluster(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks cluster

    Attributes
    ----------

    access_controls:
        List of access controls
    apply_policy_default_values:
        Whether to use policy default values for missing cluster attributes.
    autoscale:
        Autoscale specifications
    autotermination_minutes:
        Automatically terminate the cluster after being inactive for this time
        in minutes.
    cluster_id:
        Cluster ID. Used mostly when assigning a cluster to a job task.
    custom_tags:
        Additional tags for cluster resources. Databricks will tag all cluster
        resources (e.g., AWS EC2 instances and EBS volumes) with these tags in
        addition to default_tags. If a custom cluster tag has the same name as
        a default cluster tag, the custom tag is prefixed with an x_ when it
        is propagated.
    data_security_mode:
        Select the security features of the cluster. Unity Catalog requires
        SINGLE_USER or USER_ISOLATION mode. If omitted, no security features
        are enabled. In the Databricks UI, this has been recently been renamed
        Access Mode and USER_ISOLATION has been renamed Shared, but use these
        terms here.
    driver_instance_pool_id:
        Similar to instance_pool_id, but for driver node. If omitted, and
        instance_pool_id is specified, then the driver will be allocated from
        that pool.
    driver_node_type_id:
        The node type of the Spark driver. This field is optional; if unset,
        API will set the driver node type to the same value as node_type_id
        defined above.
    enable_elastic_disk:
        If you don’t want to allocate a fixed number of EBS volumes at cluster
        creation time, use autoscaling local storage. With autoscaling local
        storage, Databricks monitors the amount of free disk space available
        on your cluster’s Spark workers. If a worker begins to run too low on
        disk, Databricks automatically attaches a new EBS volume to the worker
        before it runs out of disk space. EBS volumes are attached up to a
        limit of 5 TB of total disk space per instance (including the
        instance’s local storage). To scale down EBS usage, make sure you have
        autotermination_minutes and autoscale attributes set.
    enable_local_disk_encryption:
        Some instance types you use to run clusters may have locally attached
        disks. Databricks may store shuffle data or temporary data on these
        locally attached disks. To ensure that all data at rest is encrypted
        for all storage types, including shuffle data stored temporarily on
        your cluster’s local disks, you can enable local disk encryption. When
        local disk encryption is enabled, Databricks generates an encryption
        key locally unique to each cluster node and uses it to encrypt all data
        stored on local disks. The scope of the key is local to each cluster
        node and is destroyed along with the cluster node itself. During its
        lifetime, the key resides in memory for encryption and decryption and
        is stored encrypted on the disk. Your workloads may run more slowly
        because of the performance impact of reading and writing encrypted
        data to and from local volumes. This feature is not available for all
        Azure Databricks subscriptions. Contact your Microsoft or Databricks
        account representative to request access.
    idempotency_token:
        An optional token to guarantee the idempotency of cluster creation
        requests. If an active cluster with the provided token already exists,
        the request will not create a new cluster, but it will return the
        existing running cluster's ID instead. If you specify the idempotency
        token, upon failure, you can retry until the request succeeds.
        Databricks platform guarantees to launch exactly one cluster with that
        idempotency token. This token should have at most 64 characters.
    init_scripts:
        List of init scripts specifications
    instance_pool_id:
        To reduce cluster start time, you can attach a cluster to a predefined
        pool of idle instances. When attached to a pool, a cluster allocates
        its driver and worker nodes from the pool. If the pool does not have
        sufficient idle resources to accommodate the cluster’s request, it
        expands by allocating new instances from the instance provider. When
        an attached cluster changes its state to TERMINATED, the instances it
        used are returned to the pool and reused by a different cluster.
    is_pinned:
        boolean value specifying if the cluster is pinned (not pinned by
        default). You must be a Databricks administrator to use this. The
        pinned clusters' maximum number is limited to 100, so apply may fail
        if you have more than that (this number may change over time, so check
        Databricks documentation for actual number).
    libraries:
        List of libraries specifications
    name:
        Cluster name, which doesn’t have to be unique. If not specified at
        creation, the cluster name will be an empty string.
    node_type_id:
        Any supported databricks.getNodeType id. If instance_pool_id is
        specified, this field is not needed.
    num_workers:
        Number of worker nodes that this cluster should have. A cluster has one Spark driver and num_workers executors
        for a total of num_workers + 1 Spark nodes.
    policy_id:
    runtime_engine:
        The type of runtime engine to use. If not specified, the runtime engine type is inferred
        based on the spark_version value
    single_user_name:
        The optional user name of the user to assign to an interactive cluster. This field is required when using
        data_security_mode set to SINGLE_USER or AAD Passthrough for Azure Data Lake Storage (ADLS) with a single-user
        cluster (i.e., not high-concurrency clusters).
    spark_conf:
        Map with key-value pairs to fine-tune Spark clusters, where you can provide custom Spark configuration
        properties in a cluster configuration.
    spark_env_vars:
        Map with environment variable key-value pairs to fine-tune Spark clusters. Key-value pairs of the form (X,Y)
        are exported (i.e., X='Y') while launching the driver and workers.
    spark_version:
        Runtime version of the cluster. Any supported databricks.getSparkVersion id. We advise using Cluster Policies
        to restrict the list of versions for simplicity while maintaining enough control.
    ssh_public_keys:
        SSH public key contents that will be added to each Spark node in this cluster. The corresponding private keys
        can be used to login with the user name ubuntu on port 2200. You can specify up to 10 keys.


    Examples
    --------
    ```py
    from laktory import models

    cluster = models.Cluster(
        name="default",
        spark_version="14.0.x-scala2.12",
        data_security_mode="USER_ISOLATION",
        node_type_id="Standard_DS3_v2",
        autoscale={
            "min_workers": 1,
            "max_workers": 4,
        },
        num_workers=0,
        autotermination_minutes=30,
        libraries=[{"pypi": {"package": "laktory==0.0.23"}}],
        access_controls=[
            {
                "group_name": "role-engineers",
                "permission_level": "CAN_RESTART",
            }
        ],
        is_pinned=True,
    )
    ```

    References
    ----------

    * [Databricks Cluster](https://docs.databricks.com/en/compute/configure.html#autoscaling-local-storage-1)
    * [Pulumi Databricks Cluster](https://www.pulumi.com/registry/packages/databricks/api-docs/cluster/)

    """

    access_controls: list[AccessControl] = []
    apply_policy_default_values: bool = None
    autoscale: ClusterAutoScale = None
    autotermination_minutes: int = None
    # aws_attributes
    # azure_attributes
    cluster_id: str = None
    # cluster_log_conf
    # cluster_source
    # cluster_mount_infos
    custom_tags: dict[str, str] = None
    data_security_mode: Literal[
        "NONE", "SINGLE_USER", "USER_ISOLATION"
    ] = "USER_ISOLATION"
    # docker_image
    driver_instance_pool_id: str = None
    driver_node_type_id: str = None
    enable_elastic_disk: bool = None
    enable_local_disk_encryption: bool = None
    # gcp_attributes
    idempotency_token: str = None
    init_scripts: list[ClusterInitScript] = []
    instance_pool_id: str = None
    is_pinned: bool = True
    libraries: list[ClusterLibrary] = []
    name: str = None
    node_type_id: str = Field(...)
    num_workers: int = None
    policy_id: str = None
    runtime_engine: Literal["STANDARD", "PHOTON"] = None
    single_user_name: str = None
    spark_conf: dict[str, str] = {}
    spark_env_vars: dict[str, str] = {}
    spark_version: str
    ssh_public_keys: list[str] = []
    # workload_type:

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    cluster_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Cluster"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Cluster

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"name": "cluster_name"}

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "init_scripts": "init_scripts",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_cluster"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import subprocess


class Worker:
    def run(self, cmd, cwd=None, raise_exceptions=True):
        try:
            completed_process = subprocess.run(
                cmd,
                cwd=cwd,
                check=True,
            )

        except Exception as e:
            if raise_exceptions:
                raise e
            else:
                print("An error occurred:", str(e))



import pytest
from laktory import dlt


def test_dlt():
    assert dlt.is_mocked()
    assert dlt.is_debug()

    assert hasattr(dlt, "table")
    assert hasattr(dlt, "read")
    assert hasattr(dlt, "read_stream")


if __name__ == "__main__":
    test_dlt()



from pydantic import field_validator
from pydantic import model_validator
from typing import Union
from typing import Any
from typing import Callable

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.spark import DataFrame
from laktory.spark import Column as SparkColumn
from laktory.models.sql.column import Column

logger = get_logger(__name__)


class TimeWindow(BaseModel):
    """
    Specifications for Time Window Aggregation

    Attributes
    ----------
    time_column:
        Timestamp column used for grouping rows
    window_duration:
        Duration of the window e.g. ‘1 second’, ‘1 day 12 hours’, ‘2 minutes’
    slide_duration
        Duration of the slide. If a slide is smaller than the window, windows
        are overlapping
    start_time
        Offset with respect to 1970-01-01 00:00:00 UTC with which to start
        window intervals.

    References
    ----------

    * [spark structured streaming windows](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#types-of-time-windows)
    * [pyspark window](https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.window.html)

    """

    time_column: str
    window_duration: str
    slide_duration: Union[str, None] = None
    start_time: Union[str, None] = None


# TODO: Add support for partition by and drop
"""
w = Window.partitionBy("account_id").orderBy(F.col("created_at").desc())
df = df.withColumn("_row", F.row_number().over(w)).filter(F.col("_row") == 1)
df = df.drop("_row")
"""


class TableAggregation(BaseModel):
    """
    Specifications for aggregation by group or time window.

    Attributes
    ----------
    groupby_window:
        Aggregation window definition
    groupby_columns:
        List of column names to group by
    agg_expressions:
        List of columns defining the aggregations

    Examples
    --------
    ```py
    from laktory import models

    table = models.Table(
        name="gld_stock_prices_by_1d",
        builder={
            "layer": "GOLD",
            "table_source": {
                "name": "slv_star_stock_prices",
            },
            "aggregation": {
                "groupby_columns": ["symbol"],
                "groupby_window": {
                    "time_column": "_tstamp",
                    "window_duration": "1 day",
                },
                "agg_expressions": [
                    {"name": "low", "spark_func_name": "min", "spark_func_args": ["low"]},
                    {"name": "high", "spark_func_name": "max", "spark_func_args": ["high"]},
                ],
            },
        },
    )
    ```


    References
    ----------

    * [pyspark window](https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.window.html)
    """

    groupby_window: Union[TimeWindow, None] = None
    groupby_columns: Union[list[str], None] = []
    agg_expressions: list[Column] = []

    @model_validator(mode="after")
    def groupby(self) -> Any:
        if self.groupby_window is None and len(self.groupby_columns) == 0:
            raise ValueError(
                "Either `groupby_window` or `groupby_columns` must be specified."
            )

        return self

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def execute(
        self,
        df,
        udfs: list[Callable[[...], SparkColumn]] = None,
    ) -> DataFrame:
        import pyspark.sql.functions as F

        # Groupby arguments
        groupby = []
        if self.groupby_window:
            groupby += [
                F.window(
                    timeColumn=self.groupby_window.time_column,
                    windowDuration=self.groupby_window.window_duration,
                    slideDuration=self.groupby_window.slide_duration,
                    startTime=self.groupby_window.start_time,
                )
            ]

        for c in self.groupby_columns:
            groupby += [c]

        # Agg arguments
        aggs = []
        for expr in self.agg_expressions:
            expr.type = "_any"
            aggs += [
                expr.to_spark(
                    df=df,
                    udfs=udfs,
                    raise_exception=True,
                ).alias(expr.name)
            ]

        return df.groupby(groupby).agg(*aggs)


if __name__ == "__main__":
    from laktory import models

    table = models.Table(
        name="gld_stock_prices_by_1d",
        builder={
            "layer": "GOLD",
            "table_source": {
                "name": "slv_star_stock_prices",
            },
            "aggregation": {
                "groupby_columns": ["symbol"],
                "groupby_window": {
                    "time_column": "_tstamp",
                    "window_duration": "1 day",
                },
                "agg_expressions": [
                    {
                        "name": "low",
                        "spark_func_name": "min",
                        "spark_func_args": ["low"],
                    },
                    {
                        "name": "high",
                        "spark_func_name": "max",
                        "spark_func_args": ["high"],
                    },
                ],
            },
        },
    )



from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class SecretAcl(BaseModel, PulumiResource):
    """
    Databricks secret ACL

    Attributes
    ----------
    permission:
        Scope associated with the secret
    principal:
        Key associated with the secret.
    scope:
        Value associated with the secret
    """

    permission: str = None
    principal: str = None
    scope: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return f"{self.scope}-{self.principal}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SecretAcl"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.SecretAcl

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_secret_acl"



from laktory.models.stacks.stack import Stack
from laktory.models.stacks.stack import StackResources
from laktory.models.stacks.pulumistack import PulumiStack
from laktory.models.stacks.terraformstack import TerraformStack
